/**
 * Copyright (C) 2007 4D Soft Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * See the LICENSE file for the full license
 */

package hu.soft4d.minerva.ssm;

import hu.soft4d.minerva.uml.UmlClass;
import hu.soft4d.minerva.uml.UmlEnumeration;
import hu.soft4d.minerva.uml.UmlOperation;
import hu.soft4d.minerva.uml.UmlParameter;
import hu.soft4d.minerva.uml.UmlState;
import hu.soft4d.minerva.uml.UmlStateMachine;
import hu.soft4d.minerva.uml.UmlTransition;
import hu.soft4d.minerva.uml.UmlType;
import hu.soft4d.minerva.uml.xmi.XmiUMLSSMFactory;
import hu.soft4d.minerva.wsdl.ssm.SSMHelper;
import hu.soft4d.minerva.wsdl.ssm.WSDLSSMFactory;
import hu.soft4d.util.CollectionUtil;

import java.io.PrintStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.namespace.QName;

import org.apache.log4j.Logger;

import stsimulator.sts.InteractionVariable;
import stsimulator.sts.Location;
import stsimulator.sts.LocationVariable;
import stsimulator.sts.Message;
import stsimulator.sts.MessageKind;
import stsimulator.sts.Operation;
import stsimulator.sts.OperationKind;
import stsimulator.sts.Port;
import stsimulator.sts.Service;
import stsimulator.sts.Switch;
import stsimulator.sts.SymbolicTransitionSystem;
import stsimulator.types.ComplexType;
import stsimulator.types.Enumeration;
import stsimulator.types.Type;

/**
 * Factory class for generating a SymbolicTransitionSystem, if a WSDL/SSM schema
 * definition URL or an UML definition is given.
 * 
 * <p>Intended usage of this class is:
 *  <pre>
 *    SSMFactory ssmf = SSMFactory.getFactoryFromWSDL(URL wsdlURL, URL spssmURL, 
 *                               String serviceName, String portName);
 *    // or
 *    SSMFactory ssmf = SSMFactory.getFactoryFromUML(String modelFileName, String ssmName);
 *    
 *    SymbolicTransitionSystem ssm = ssmf.getSSM();
 *  </pre>
 * 
 * <p> Known subclasses are: WSDLSSMFactory and XmiUMLSSMFactory.
 * 
 * @author Zsolt Kiss Gere, 4D Soft Ltd.
 */
public abstract class SSMFactory
{
    static Logger log = Logger.getLogger("SSMFactory");
    
    protected String wsdlUri;
    protected String namespace;
    protected String serviceName;
    protected String portName;
    protected Service service;
    protected Port port;
    // The name of the initial state
    protected String initialState;
    
    // SSM containers
    // Locations (states) indexed by name
    protected Map<String, Location> locationMap = new HashMap<String, Location>();
    protected List<Switch> switchList = new ArrayList<Switch>();
    protected Map<String, LocationVariable> locvarMap = new HashMap<String, LocationVariable>();
    protected Map<String, InteractionVariable> intvarMap = new HashMap<String, InteractionVariable>();
    protected Map<String, Message> messageMap = new HashMap<String, Message>();
    protected Map<String, Operation> operationMap = new HashMap<String, Operation>();
    protected Map<QName, Type> typeMap = new HashMap<QName, Type>();
    
    private SymbolicTransitionSystem ssm;

    public static SSMFactory getFactoryFromWSDL( String wsdlUri, String spssmUri, 
                               String serviceName, String portName) throws Exception
    {
        SSMFactory ssmf = new WSDLSSMFactory(wsdlUri, spssmUri, serviceName, portName);
        return ssmf;
    }

    public static SSMFactory getFactoryFromUML(String modelFileName, String ssmName) throws Exception
    {
        SSMFactory ssmf = new XmiUMLSSMFactory( modelFileName, ssmName );
        return ssmf;
    }

    public SymbolicTransitionSystem getSSM() throws Exception
    {
        if( ssm != null )  return ssm;
        
        ssm = new SymbolicTransitionSystem();
        
        ssm.addService( service );
        
        for( Type type: typeMap.values() )
            ssm.addType( type );
        
        for( LocationVariable var: locvarMap.values() )
            ssm.addVariable( var );
        
        for( Location location: locationMap.values() )
            ssm.addLocation( location );
        
        for( Switch sw: switchList )
            ssm.addSwitch( sw );

        for( InteractionVariable var: intvarMap.values() )
            ssm.addVariable( var );
        
        for( Operation oper: operationMap.values() )
            ssm.addOperation( oper );
        
        for( Message msg: messageMap.values() )
            ssm.addMessage( msg );
        
        if( initialState != null )
        {
            Location initialLoc = locationMap.get( initialState );
            ssm.setCurrentLocation( initialLoc );
        }
        
        return ssm;
    }
    
    /**
     * Fill the containers from an UmlStateMachine.
     */
    protected void initFromStateMachine( UmlStateMachine mystm ) throws Exception
    {
        wsdlUri = mystm.getWsdlUrl();
        namespace = mystm.getNamespace();
        serviceName = mystm.getService();
        portName = mystm.getPort();
        
        port = new Port( portName );
        service = new Service( serviceName );
        service.addPort( port );
        
        // The location variables are the global variables of the SSM
        // They are represented as attributes (ownedAttribute-s) of the UML SSM object
        UmlParameter[] locvars = mystm.getAttributes();
        for( UmlParameter locvar: locvars )
        {
            String name = locvar.getName();
            // Get the SSM type corresponding to the parameter type. typeMap is from SSMFactory
            Type type = getCreateSSMType(locvar, typeMap);
            
            // The following call fills the locvarMap
            getCreateLocationVariable(name, type);
        }
        
        // Locations - the states of the ssm
        UmlState[] states = mystm.getStates();
        for( UmlState state: states )
        {
            String name = state.getName();
            locationMap.put(name, new Location(name));
            
            if( state.isInitialState() )
                if( initialState == null )
                    initialState = name;
                else
                    throw new Exception("Initial state already set to " + initialState );
        }
//        
//        // Switches - transitions
//        UmlTransition[] transitions = mystm.getTransitions();
//        for( UmlTransition trans: transitions )
//        {
//            Switch sswitch = createSwitch( trans );
//            switchList.add( sswitch );
//        }
    }

    /**
     *  Check if an ssm operation is stored under the key name. 
     *  If yes, return it.
     *  Otherwise create a new operation, store and return it.
     */
    protected Operation getCreateSSMOperation(String name, OperationKind kind)
    {
        Operation oper = operationMap.get(name);
        if( oper == null )
        {
            oper = new Operation(name, kind);
            //System.out.println("Created new Operation <" + name + "> of kind " + kind );
            
            operationMap.put(name, oper);
        }
        return oper;
    }
    
    /**
     *  If a location variable with the given name exists in the locvarMap,
     *  return it. Otherwise create a new location variable, record it in the map 
     *  and return it. 
     */
    protected LocationVariable getCreateLocationVariable(String name, Type type)
    {
        String key = name;
        LocationVariable var = locvarMap.get(key);
        if( var == null )
        {
            var = new LocationVariable(name, type, type.generateInitialInstance());
            locvarMap.put(key, var);
        }
        return var;
    }

    /**
     *  If an interaction variable with the given (name, type) exists in the intvarMap,
     *  return it. Otherwise create a new interaction variable, record it in the map 
     *  and return it. 
     */
    protected InteractionVariable getCreateInteractionVariable(String name, Type type)
    {
        String key = getVarKey(name, type);
        InteractionVariable var = intvarMap.get(key);
        if( var == null )
        {
            var = new InteractionVariable(name, type);
            intvarMap.put(key, var);
        }
        return var;
    }
    
    /**
     *  Get the SSM type corresponding to the type of the UmlParameter parameter.
     *  Recursive method. Uses typeMap for cache.
     */
    protected Type getCreateSSMType(UmlParameter parameter, Map<QName, Type> typeMap) throws Exception
    {
        UmlType umltype = parameter.getUmlType();
        String typeName = umltype.getName();
        QName typeQName = new QName(namespace, typeName); 
        
        // If the type is found in the map, return it
        Type ssmType = typeMap.get( typeQName );
        if( ssmType != null )
            return ssmType;
        
        // Type is an enumeration
        if( umltype instanceof UmlEnumeration )
        {
            ssmType = new Enumeration( typeQName, ((UmlEnumeration) umltype).getLiterals() );
            typeMap.put(typeQName, ssmType);
            return ssmType;
        }
        
        // A new simple type (String, int, ...), not in the map
        // NOTE: The UML specification declares 'int' as a DataType and String as a PrimitiveType, 
        // which is derived from DataType. But 'int' is not PrimitiveType.
        else if( umltype.isSimpleType() )
        {
            // TODO !! CHECK IF the type names match the ones in getSSMSimpleType()!!
            ssmType = SSMHelper.getSSMSimpleType( typeName );
            typeMap.put(typeQName, ssmType);
            return ssmType;
        }
        
        // Class
        else if( umltype instanceof UmlClass )
        {
            ComplexType complex = new ComplexType( typeQName );
            typeMap.put(typeQName, complex);
            
            // Get the parameters of our complex type
            UmlParameter[] attributes = ((UmlClass) umltype).getAttributes();
            Object[][] parArray = new Object[attributes.length][2];
            int i = 0;
            for( UmlParameter attr: attributes )
            {
                String name = attr.getName();
                // Now call recursively
                Type type = getCreateSSMType( attr, typeMap );
                // We are ready now to construct an SSM parameter with (name, parSSMType)
                parArray[i][0] = name;
                parArray[i][1] = type;
                i++;
            }
            complex.setFields( parArray );
            
            return complex;
        }
        
        else throw new Exception("Unknown type: " + umltype.getName());
    }
    
    /**
     *  Creates an sts Switch from an UmlTransition. An sts message, with all its 
     *  dependencies is also created if needed. 
     *  
     *  <p>The following containers are filled/used when creating a switch:
     *  messageMap, ssmTypes, intvarMap and operationMap.
     *  
     *  <p>locationMap is only used in read mode. 
     */
    protected Switch createSwitch( UmlTransition trans ) throws Exception
    {
        // To create a switch, we need a message, a restriction and an update.
        // The Message
        Message ssmMsg = null;
        
        // If the transition kind is unobservable, the message is also the like. 
        if( trans.getSSMKind() == UmlTransition.Kind.UNOBSERVABLE )
        {
            ssmMsg = Message.generateUnobservableMessageForPort(port);
            // Can one unobservable message be used more times in operations?
            messageMap.put("", ssmMsg);
        }
        else {
            // An (operation, kind) identifies a(n ssm) message
            // which could have been already stored
            
            // Get the operation
            UmlOperation operation = trans.getOperation();
            
            String msgKey = operation.getName() + "@" + trans.getSSMKind().toString().toLowerCase();
            
            ssmMsg = messageMap.get(msgKey);
            if( ssmMsg == null )
            {
                //log.info("Creating a message: " + msgKey );
                ssmMsg = createSSMMessage( trans, operation );
                messageMap.put(msgKey, ssmMsg);
            }
            else {
                //log.info("Got message with key: " + msgKey );
            }
        }
        
        String restriction = trans.getGuard();
        String update = trans.getUpdateCommand();
        
        Switch sswitch = new Switch( ssmMsg, restriction, update );
        
        // Insert switch - location associations
        Location startLoc = locationMap.get( trans.getSource().getName() );
        startLoc.addOutgoingSwitch( sswitch );
        
        Location endLoc = locationMap.get( trans.getTarget().getName() );
        sswitch.setTarget( endLoc );
        
        return sswitch;
    }

    /**
     *  Containers used in read/write mode: ssmTypes, intvarMap and operationMap
     */
    protected Message createSSMMessage( UmlTransition trans, UmlOperation operation ) throws Exception
    {
        String msgName = null;
        MessageKind msgKind = null;
        UmlParameter[] parameters = null;
        
        if( trans.getSSMKind() == UmlTransition.Kind.INPUT )
        {
            msgName = operation.getInputMessageName();
            msgKind = MessageKind.INPUT;
            
            // The message parameters are the parameters of the operation
            parameters = operation.getParameters();
        }
        else if( trans.getSSMKind() == UmlTransition.Kind.OUTPUT )
        {
            msgName = operation.getOutputMessageName();
            msgKind = MessageKind.OUTPUT;
            
            // There is one message parameter: the return parameter of the operation
            parameters = new UmlParameter[] {operation.getReturnParameter()};
        }
        // If neither tag exists, fall back to the operation name + "Response" for OUTPUT msgs
        if( msgName == null )
        {
            if( trans.getSSMKind() == UmlTransition.Kind.OUTPUT )
                msgName = operation.getName() + "Response";
            else
                msgName = operation.getName();
        }
        
        // The SSM parameter list
        ArrayList<InteractionVariable> paramList = new ArrayList<InteractionVariable>();
        
        // For each parameter, create or retrieve an interaction variable and save it
        for( UmlParameter par: parameters )
        {
            String name = par.getName();
            
            // Get the SSM type corresponding to the parameter type
            Type type = getCreateSSMType(par, typeMap);
            
            InteractionVariable var = getCreateInteractionVariable(name, type);
            paramList.add( var );
        }
        
        Operation oper = getCreateSSMOperation(operation.getName(), operation.getOperationKind());
        
        Message ssmMsg = new Message( msgName, msgKind, oper, port, paramList );
        return ssmMsg;
    }

    protected String getVarKey(String name, Type type)
    {
        return name + "@" + type.getName(); 
    }
    
    public void dumpSSM(PrintStream out)
    {
        out.println("==== DUMP BEGIN ==================================");
        out.println();

        out.println("wsdlUri: <<" + wsdlUri + ">>");
        out.println("namespace: <<" + namespace + ">>");
        out.println("serviceName: <<" + serviceName + ">>");
        out.println("portName: <<" + portName + ">>");
        out.println("initialState: <<" + initialState + ">>");
        out.println();
        
        CollectionUtil.dumpMap(out, "Types", typeMap);
        CollectionUtil.dumpMap(out, "Locations", locationMap);
        CollectionUtil.dumpMap(out, "Location variables", locvarMap);
        CollectionUtil.dumpMap(out, "Interaction variables", intvarMap);
        CollectionUtil.dumpMap(out, "Operations", operationMap);
        CollectionUtil.dumpMap(out, "Messages", messageMap);
        // Dump switchList
        CollectionUtil.dumpSwitches(out, "Switches", switchList);
        
        out.println();
        out.println("==== DUMP END ====================================");
    }
    public String getInitialState()
    {
        return initialState;
    }

    public Map<String, InteractionVariable> getIntvarMap()
    {
        return intvarMap;
    }

    public Map<String, Location> getLocationMap()
    {
        return locationMap;
    }

    public Map<String, LocationVariable> getLocvarMap()
    {
        return locvarMap;
    }

    public Map<String, Message> getMessageMap()
    {
        return messageMap;
    }

    public String getNamespace()
    {
        return namespace;
    }

    public Map<String, Operation> getOperationMap()
    {
        return operationMap;
    }

    public Port getPort()
    {
        return port;
    }

    public String getPortName()
    {
        return portName;
    }

    public Service getService()
    {
        return service;
    }

    public String getServiceName()
    {
        return serviceName;
    }

    public List<Switch> getSwitchList()
    {
        return switchList;
    }

    public Map<QName, Type> getTypeMap()
    {
        return typeMap;
    }

    public String getWsdlUri()
    {
        return wsdlUri;
    }
}
