/**
 * Copyright (C) 2007 4D Soft Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * See the LICENSE file for the full license
 */

package hu.soft4d.minerva.uml.xmi;

import hu.soft4d.minerva.ssm.SSMFactory;
import hu.soft4d.minerva.uml.UmlParameter;
import hu.soft4d.minerva.wsdl.ssm.SSMHelper;

import stsimulator.sts.*;
import stsimulator.types.*;

import java.util.ArrayList;
import java.util.Map;
import java.util.logging.Logger;

import javax.xml.namespace.QName;

/**
 * @author Zsolt Kiss Gere, 4D Soft Ltd.
 */
public class XmiUMLSSMFactory extends SSMFactory
{
    private static Logger log = Logger.getLogger("XmiUMLSSMFactory");
    
    String modelFileName;
    String ssmName;
    XmiUmlModel model;
    
    public XmiUMLSSMFactory( String modelFileName, String ssmName ) throws Exception
    {
        this.modelFileName = modelFileName;
        this.ssmName = ssmName;
        
        model = new XmiUmlModel( modelFileName );
        initialize();
    }

    /**
     * Fill the containers.
     */
    private void initialize() throws Exception
    {
        XmiUmlStateMachine usm = model.getUmlStateMachine( ssmName );
        if( usm == null )
            throw new Exception("StateMachine '" + ssmName + "' not found");
        
        //log.info("Got usm: " + usm);
        
        wsdlUri = usm.wsdlUrl;
        namespace = usm.namespace;
        serviceName = usm.service;
        portName = usm.port;
        
        port = new Port( portName );
        service = new Service( serviceName );
        service.addPort( port );
        
        // The location variables are the global variables of the SSM
        // They are represented as attributes (ownedAttribute-s) of the UML SSM object
        XmiUmlParameter[] locvars = usm.attributes;
        for( XmiUmlParameter locvar: locvars )
        {
            String name = locvar.name();

            // Get the SSM type corresponding to the parameter type. typeMap is from SSMFactory
            Type type = getCreateSSMType(locvar, typeMap);
            
            // The following call fills the locvarMap
            LocationVariable var = getCreateLocationVariable(name, type);
        }
        
        // Locations - the states of the ssm
        XmiUmlState[] states = usm.states;
        for( XmiUmlState state: states )
        {
            String name = state.name();
            locationMap.put(name, new Location(name));
            
            if( state.isInitialState() )
                if( initialState == null )
                    initialState = name;
                else
                    throw new Exception("Initial state already set to " + initialState );
        }
        
        // Switches - transitions
        XmiUmlTransition[] transitions = usm.transitions;
        for( XmiUmlTransition trans: transitions )
        {
            Switch sswitch = createSwitch( trans );
            switchList.add( sswitch );
        }
    }

    /**
     *  The following containers are filled/used when creating a switch:
     *  messageMap, ssmTypes, intvarMap and operationMap.
     *  
     *  <p>locationMap is only used in read mode. 
     */
    private Switch createSwitch( XmiUmlTransition trans ) throws Exception
    {
        // To create a switch, we need a message, a restriction and an update.
        
        // The Message
        Message ssmMsg = null;
        
        // If the transition kind is unobservable, the message is also the like. 
        if( trans.kind == XmiUmlTransition.Kind.UNOBSERVABLE )
        {
            ssmMsg = Message.generateUnobservableMessageForPort(port);
            // Can one unobservable message be used more times in operations?
            messageMap.put("", ssmMsg);
        }
        else {
            // An (operation, kind) identifies a(n ssm) message
            // which could have been already stored
            
            // Get the operation
            XmiUmlCallTrigger trigger = trans.getCallTrigger();
            XmiUmlOperation operation = trigger.getOperation();
            
            if( operation == null )
                throw new Exception("Found trigger with no operation - not allowed. " +
                        "\n  transition " + trans.getSourceNode().name() + " -> " + trans.getTargetNode().name() +
                        "\n  Check 'deleted' transitions");
            
            log.info("transition: " + trans);
            
            String msgKey = operation.name() + "@" + trans.kind.toString().toLowerCase();
            
            ssmMsg = messageMap.get(msgKey);
            if( ssmMsg == null )
            {
                log.info("Creating a message: " + msgKey );
                
                ssmMsg = createSSMMessage( trans, operation );
                messageMap.put(msgKey, ssmMsg);
            }
            else {
                log.info("Got message with key: " + msgKey );
            }
        }
        
        String restriction = trans.guard.toString();
        String update = trans.update;
        
        Switch sswitch = new Switch( ssmMsg, restriction, update );
        
        // Insert switch - location associations
        Location startLoc = locationMap.get( trans.getSourceNode().name() );
        startLoc.addOutgoingSwitch( sswitch );
        
        Location endLoc = locationMap.get( trans.getTargetNode().name() );
        sswitch.setTarget( endLoc );
        
        return sswitch;
    }

    /**
     *  Containers used in read/write mode: ssmTypes, intvarMap and operationMap
     */
    private Message createSSMMessage( XmiUmlTransition trans, XmiUmlOperation operation ) throws Exception
    {
        // The message name is given as tags "inputMessage" or "outputMessage" of 
        // <<WebMethod>>
        String msgName = null;
        MessageKind msgKind = null;
        XmiUmlParameter[] parameters = null;
        
        if( trans.kind == XmiUmlTransition.Kind.INPUT )
        {
            msgName = operation.getInputMessageName();
            msgKind = MessageKind.INPUT;
            // The message parameters are the parameters of the operation
            parameters = operation.params;
        }
        else if( trans.kind == XmiUmlTransition.Kind.OUTPUT )
        {
            msgName = operation.getOutputMessageName();
            msgKind = MessageKind.OUTPUT;
            // There is one message parameter: the return parameter of the operation
            parameters = new XmiUmlParameter[] {operation.retParam};
        }
        // If neither tag exists, fall back to the operation name + "Response" for OUTPUT msgs
        if( msgName == null )
        {
            if( trans.kind == XmiUmlTransition.Kind.OUTPUT )
                msgName = operation.name() + "Response";
            else
                msgName = operation.name();
        }
        
        // The SSM parameter list
        ArrayList<InteractionVariable> paramList = new ArrayList<InteractionVariable>();
        
        // For each parameter, create or retrieve an interaction variable and save it
        for( XmiUmlParameter par: parameters )
            if( par != null )
            {
                String name = par.name();
                
                // Get the SSM type corresponding to the parameter type
                Type type = getCreateSSMType(par, typeMap);
                
                InteractionVariable var = getCreateInteractionVariable(name, type);
                paramList.add( var );
            }
        
        Operation oper = getCreateSSMOperation(operation.name(), operation.getOperationKind());
        
        Message ssmMsg = new Message( msgName, msgKind, oper, port, paramList );
        return ssmMsg;
    }
    
    /**
     *  Get the SSM type corresponding to the type of the XmiUmlParameter parameter.
     *  Recursive method. Uses typeMap for cache.
     *  
     *  NOTE: Parent SSMFactory has a getCreateSSMType(UmlParameter parameter,...), but
     *        its first parameter is different.
     */
    private Type getCreateSSMType( XmiUmlParameter parameter, Map<QName,Type> typeMap ) throws Exception
    {
        String typeName = parameter.getTypeName();
        QName typeQName = new QName(namespace, typeName); 
        
        // If the type is found in the map, return it
        Type ssmType = typeMap.get( typeQName );
        if( ssmType != null )
            return ssmType;
        
        // A new primitive type, not in the map
        if( parameter.hasSimpleType() )
        {
            ssmType = SSMHelper.getSSMSimpleType( typeName );
            typeMap.put(typeQName, ssmType);
            return ssmType;
        }
        
        // An enumeration or a complex type
        XmiUmlNode typeNode = parameter.getTypeNode();
        
        if( typeNode instanceof XmiUmlEnumeration )
        {
            ssmType = new Enumeration(typeQName, ((XmiUmlEnumeration) typeNode).values );
            typeMap.put(typeQName, ssmType);
            return ssmType;
        }
        
        // Class
        else if( typeNode instanceof XmiUmlClass )
        {
            ComplexType complex = new ComplexType( typeQName );
            typeMap.put(typeQName, complex);
            
            // Get the parameters of our complex type
            XmiUmlParameter[] attributes = ((XmiUmlClass) typeNode).attributes;
            Object[][] parArray = new Object[attributes.length][2];
            int i = 0;
            for( XmiUmlParameter attr: attributes )
            {
                String name = attr.name();
                // Now call recursively
                Type type = getCreateSSMType( attr, typeMap );
                // We are ready now to construct an SSM parameter with (name, parSSMType)
                parArray[i][0] = name;
                parArray[i][1] = type;
                i++;
            }
            complex.setFields( parArray );
            
            return complex;
        }
        
        else throw new Exception("Unknown node type: " + typeNode);
    }
}




