/**
 * Copyright (C) 2007 4D Soft Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * See the LICENSE file for the full license
 */

package hu.soft4d.minerva.uml.xmi;

import hu.soft4d.minerva.uml.xmi.XmiUmlParameter.ParameterKind;
import hu.soft4d.minerva.uml.xmi.XmiUmlState.StateKind;
import hu.soft4d.minerva.wsdl.Constants;
import hu.soft4d.minerva.xmi.XmiUtil;
import hu.soft4d.util.CollectionUtil;

import java.io.InputStream;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import xmitools.xmiparser.XmiElement;
import xmitools.xmiparser.XmiModel;
import xmitools.xmiparser.XmiNode;
import xmitools.xmiparser.XmiNodeIterator;

/**
 * @author Zsolt Kiss Gere, 4D Soft Ltd.
 */
public class XmiUmlModel
{
    private XmiModel model;
    private XmiElement root;
    
    // All registered UmlNode objects indexed by xmi:id
    Map<String, XmiUmlNode> nodes = new HashMap<String, XmiUmlNode>();
    
    // Enumerations indexed by name  
    private Map<String, XmiUmlEnumeration> enumerations = new HashMap<String, XmiUmlEnumeration>();
    
    // Classes indexed by name
    private Map<String, XmiUmlClass> umlClasses = new HashMap<String, XmiUmlClass>();
    
    // Interfaces indexed by name 
    private Map<String, UmlInterface> umlInterfaces = new HashMap<String, UmlInterface>();
    
    // State machines indexed by name
    private Map<String, XmiUmlStateMachine> umlStateMachines = new HashMap<String, XmiUmlStateMachine>();
    
    // States indexed by name
    private Map<String, XmiUmlState> umlStates = new HashMap<String, XmiUmlState>();
    
    /**
     * @param xmiFileName      Path of the XMI file which contains the model.
     *                           Currently tested for MagicDraw .uml2 files. 
     * @throws Exception
     */
    public XmiUmlModel( String xmiFileName ) throws Exception
    {
        System.err.println("reading " + xmiFileName);
        model = xmitools.umlparser.Main.parseModel(xmiFileName, true);
        root = model.root();
        
        // Build the model
        parse();
        
        //dump( System.out );
    }
    
    public void dump(PrintStream out)
    {
        out.println("==== DUMP BEGIN ==================================");
        out.println();
        
        //dumpMap(out, "Registered nodes", nodes);
        
        CollectionUtil.dumpMap(out, "Enumerations", enumerations);
        
        CollectionUtil.dumpMap(out, "Classes", umlClasses);
        
        CollectionUtil.dumpMap(out, "Interfaces", umlInterfaces);
        
        CollectionUtil.dumpMap(out, "State machines", umlStateMachines);
        
        out.println();
        out.println("==== DUMP END ====================================");
    }


    /**
     *  Build the model
     */
    protected void parse() throws Exception
    {
        XmiNodeIterator it = root.iterator();
        while( it.hasNext() )
        {
            XmiNode node = it.next();
            parseNode( node );
        }
    }

    /**
     *  This method calls itself recursively for all ownedMember-s of an uml:Package.
     */
    private void parseNode(XmiNode node) throws Exception
    {
        // Skip ownedComment-s...
        if( node.tagKind().equals("packageImport") )
            importPackage( node );
        
        // Skip everything which is not ownedMember
        if( ! node.tagKind().equals("ownedMember") )
            return;
        
        // Now node is an ownedMember
        String xmiType = XmiUtil.getXmiType( node );
        //System.err.println("node: " + node.name() + ", " + xmiType );
        
        if( xmiType.equals("uml:Package") )
        {
            Collection members = node.collectChildrenByTagName("ownedMember");
            for( Object obj: members )
                parseNode( (XmiNode) obj );
        }
        else if( xmiType.equals("uml:Enumeration") )
            addEnumeration(node);
        
        // DataType-s are handled like Classes
        else if (xmiType.equals("uml:Class") ||
                 xmiType.equals("uml:DataType"))
            addClass(node);
            
        else if( xmiType.equals("uml:Interface") )
            addInterface(node);
        
        else if( xmiType.equals("uml:StateMachine") )
            addStateMachine(node);
    }

    private void addStateMachine(XmiNode node) throws Exception
    {
        // Check if this is a JSSM state machine. Otherwise we are not interested
        if(! XmiUtil.isOfStereotype(node, Constants.SSM_PROFILE, Constants.STEREOTYPE_STATE_MACHINE) )
            return;
        
        // Get the attributes. These are the location (global) variables
        Collection attrs = node.collectChildrenByTagName("ownedAttribute");
        XmiUmlParameter[] attributes = new XmiUmlParameter[ attrs.size() ];
        int i = 0;
        for( Object attr: attrs )
        {
            XmiNode attrNode = (XmiNode) attr;
            XmiUmlParameter par = createParameterFromNode( attrNode );
            addNode(attrNode, par);
            attributes[i++] = par;
        }
        
        // Collect the call triggers
        Collection triggers = node.collectChildrenByTagName("ownedTrigger");
        XmiUmlCallTrigger[] calltriggers = new XmiUmlCallTrigger[ triggers.size() ];
        i = 0;
        
        for( Object trigger: triggers )
        {
            XmiNode xmitrigger = (XmiNode) trigger;
            
            XmiUmlCallTrigger mytrigger = new XmiUmlCallTrigger(this, xmitrigger);
            addNode(xmitrigger, mytrigger);
            calltriggers[i++] = mytrigger;
        }
        
        // Now get the states. They are <subvertex>-es of a <region>
        XmiNode region = node.getChildByTagName("region");
        if( region == null )
            throw new Exception("Expected tag 'region' not found");
            
        Collection<XmiNode> subvertexes = region.collectChildrenByTagName("subvertex");
        List<XmiUmlState> statelist = new ArrayList<XmiUmlState>();
        
        for( XmiNode xstate: subvertexes )
        {
            String xmitype = getXmiType(xstate);
            StateKind kind;
            
            if( xmitype.equals("uml:Pseudostate") )
                kind = StateKind.PseudoState;
            
            else if( xmitype.equals("uml:State") )
                kind = StateKind.NormalState;
            
            else
                continue;
            
            XmiUmlState umlstate = new XmiUmlState(this, xstate, kind);
            addNode(xstate, umlstate);
            
            if( xstate.name().equals("") )
                umlStates.put("EMPTY", umlstate);
            else
                umlStates.put(xstate.name(), umlstate);
            
            statelist.add(umlstate);
        }
        XmiUmlState[] mystates = new XmiUmlState[ statelist.size() ];
        i = 0;
        for( XmiUmlState state: statelist )
        {
            mystates[i] = state;
            i++;
        }
        
        // Now collect the transitions
        Collection<XmiNode> transitions = region.collectChildrenByTagName("transition");
        XmiUmlTransition[] umltransitions = new XmiUmlTransition[ transitions.size() ];
        i = 0;
        
        for( XmiNode transnode: transitions )
        {
            String sourceid = XmiUtil.getAttribute(transnode, "source");
            String targetid = XmiUtil.getAttribute(transnode, "target");
            String triggerid = XmiUtil.getAttribute(transnode, "trigger");
            
            XmiNode guard = transnode.getChildByTagName("guard");
            XmiNode effect = transnode.getChildByTagName("effect");
            
            XmiUmlTransition trans = new XmiUmlTransition(this, transnode, sourceid, targetid, triggerid, 
                    new UmlGuard(this, guard), new UmlTransitionEffect(this, effect));
            addNode(transnode, trans);
            umltransitions[i++] = trans;
        }
        
        // Should be only UmlStateMachine
        XmiUmlStateMachine stm = new XmiUmlStateMachine( this, node, attributes, calltriggers, 
                                                   mystates, umltransitions );
        
        addNode(node, stm);
        umlStateMachines.put(node.name(), stm);
    }

    private void addInterface(XmiNode node) throws Exception
    {
        // Interface operations
        Collection ops = node.collectChildrenByTagName("ownedOperation");
        XmiUmlOperation[] umlops = new XmiUmlOperation[ ops.size() ];
        int i = 0;
        
        for (Iterator iter = ops.iterator(); iter.hasNext(); i++) {
            XmiNode op = (XmiNode) iter.next();
            
            // An operation has parameters. These behave much like the class attributes
            // "Normal" parameters have a tag of <ownedParameter>, while
            // return parameters are tagged <returnResult> !!
            Collection pars = op.collectChildrenByTagName("ownedParameter");
            XmiUmlParameter[] params = new XmiUmlParameter[ pars.size() ];
            int j = 0;
            
            // "Normal" parameters
            for (Iterator iter2 = pars.iterator(); iter2.hasNext(); j++) {
                XmiNode parnode = (XmiNode) iter2.next();
                
                // Get the id of UmlNode corresponding to the type of the parameter 
                // It can be an XmiUmlPrimitiveType, XmiUmlEnumeration, ...
                XmiUmlParameter umlparam = createParameterFromNode( parnode );
                
                addNode(parnode, umlparam);
                params[j] = umlparam;
            }
            
            // Now add the return parameter
            XmiNode returnpar = op.getChildByTagName("returnResult");
            XmiUmlParameter retparam = createParameterFromNode( returnpar );
            
            
//            for (Iterator iter2 = retpars.iterator(); iter2.hasNext(); j++) {
//                XmiNode par = (XmiNode) iter2.next();
//                
//                // Get the id of UmlNode corresponding to the type of the parameter 
//                // It can be an XmiUmlPrimitiveType, XmiUmlEnumeration, ...
//                XmiUmlParameter umlparam = createParameterFromNode( par );
//                
//                addNode(par, umlparam);
//                params[j] = umlparam;
//            }
            
            XmiUmlOperation umlop = new XmiUmlOperation(this, op, params, retparam);
            addNode(op, umlop);
            umlops[i] = umlop;
        }
        
        UmlInterface intfce = new UmlInterface(this, node, umlops);
        
        addNode(node, intfce);
        umlInterfaces.put(node.name(), intfce);
    }

    private XmiUmlParameter createParameterFromNode(XmiNode node) throws Exception
    {
        if( node == null )  return null;
        
        // The node has either a "type" attribute (complex type), or a <type> child
        // (for a primitive type)
        String typeXmiId = XmiUtil.getTypeAttribute(node);
        
        if( typeXmiId != null && ! typeXmiId.equals("") )
            // This is a COMPLEX type parameter
            return new XmiUmlParameter(this, node, typeXmiId, ParameterKind.COMPLEX);
        
        else {
            // No "type" attribute, try a <type> child
            XmiNode typeChild = node.getChildByTagName("type");
            
            if(! XmiUtil.getXmiType(typeChild).equals("uml:PrimitiveType"))
                throw new Exception("Sorry, only uml:PrimitiveType types handled at this stage");
            
            // This is a PRIMITIVE type parameter
            typeXmiId = getPrimitiveTypeId( typeChild );
            
            return new XmiUmlParameter(this, node, typeXmiId, ParameterKind.PRIMITIVE);
        }
    }
    

    private void addClass(XmiNode node) throws Exception
    {
        // Class attributes
        Collection attrs = node.collectChildrenByTagName("ownedAttribute");
        XmiUmlParameter[] classAttrs = new XmiUmlParameter[ attrs.size() ];
        int i = 0;
        
        for (Iterator iter = attrs.iterator(); iter.hasNext(); i++) {
            XmiNode attr = (XmiNode) iter.next();
            //System.err.println("\t attr: " + attr.name());
            
            // Get the id of UmlNode corresponding to the type of the attribute 
            // It can be an XmiUmlPrimitiveType, XmiUmlEnumeration, ...
            XmiUmlParameter umlattr = createParameterFromNode( attr );
            
            addNode(attr, umlattr);
            classAttrs[i] = umlattr;
        }
        
        XmiUmlClass umlclass = new XmiUmlClass(this, node, classAttrs);
        
        addNode(node, umlclass);
        umlClasses.put(node.name(), umlclass);
    }

    private void addNode(XmiNode node, XmiUmlNode nodeObject)
    {
        nodes.put(getXmiId(node), nodeObject);
    }

    /**
     * @deprecated  Currently not needed.
     */
    private String getTypeNodeId(XmiNode node) throws Exception
    {
        // The node has either a "type" attribute, or a <type> child
        // Try the value of the "type" attribute
        String typeXmiId = XmiUtil.getTypeAttribute(node);
        
        if( typeXmiId != null && ! typeXmiId.equals("") )
            return typeXmiId;
        
        else {
            // No "type" attribute, try a <type> child
            XmiNode typeChild = node.getChildByTagName("type");
            
            if( XmiUtil.getXmiType(typeChild).equals("uml:PrimitiveType") )
            {
                return getPrimitiveTypeId(typeChild);
            }
            else
                throw new Exception("Sorry, only uml:PrimitiveType types handled at this stage");
        }
    }

    private String getPrimitiveTypeId(XmiNode node) throws Exception
    {
        String typeRef = XmiUtil.getAttribute(node, "href");

        if (typeRef == null || typeRef.equals(""))
            throw new Exception("Attribute 'href' not found for node " + node.name());

        // Get path
        String path = null;
        String id = null;
        int ind = typeRef.indexOf('#');
        if (ind != -1) {
            path = typeRef.substring(0, ind);
            id = typeRef.substring(ind + 1);
            return id;
        }
        else
            throw new Exception("'#' expected and not found in href " + typeRef);
    }

    private void addEnumeration(XmiNode node)
    {
        String name = node.name();
        
        // The literals
        Collection items = node.collectChildrenByTagName("ownedLiteral");
        String[] values = new String[ items.size() ];
        
        int i = 0;
        for (Iterator iter = items.iterator(); iter.hasNext(); i++) {
            XmiNode child = (XmiNode) iter.next();
            //System.err.println("\t item: " + child.name());
            values[i] = child.name();
        }
        
        XmiUmlEnumeration enum_ = new XmiUmlEnumeration(this, node, values);
        enumerations.put(name, enum_);
        addNode(node, enum_);
    }
    
    private void importPackage(XmiNode node) throws Exception
    {
        // Check if it has an <importedPackage> tag
        XmiNode pkgnode = node.getChildByTagName("importedPackage");
        if( pkgnode != null )
        {
            String xmiType = getXmiType(pkgnode); 
            if( xmiType != null && xmiType.equals("uml:Model") )
            {
                // Get href
                String href = XmiUtil.getAttribute(pkgnode, "href");
                if( href == null || href.equals("") )
                    return;
                
                // href is of form path#id
                String path = null;
                //String id = null;
                int ind = href.indexOf('#');
                if( ind != -1 )
                {
                    path = href.substring(0, ind);
                    //id = href.substring( ind+1 );
                }
                if( path != null && (path.equals(XmiUtil.LIBRARY_JAVA_PRIMITIVE_TYPES) 
                                  || path.equals(XmiUtil.LIBRARY_UML2_PRIMITIVE_TYPES)))
                {
                    loadPrimitiveTypes( path );
                }
                // else ignore for the moment
            }
        }
    }

    private void loadPrimitiveTypes(String path) throws Exception
    {
        String filename = getResourceOf( path );
        InputStream is = this.getClass().getResourceAsStream( filename );
        XmiModel xModel = xmitools.umlparser.Main.parseModelFromStream( is, true );
        XmiElement xRoot = xModel.root();
        
        Collection items = xRoot.collectChildrenByTagName("ownedMember");
        for (Iterator iter = items.iterator(); iter.hasNext();)
        {
            XmiNode node = (XmiNode) iter.next();
            if( XmiUtil.getXmiType(node).equals("uml:PrimitiveType") )
            {
                String id = XmiUtil.getXmiId( node );
                nodes.put(id, new XmiUmlPrimitiveType(this, node));
            }
        }
    }

    private String getResourceOf(String path) throws Exception
    {
        if( path.equals(XmiUtil.LIBRARY_JAVA_PRIMITIVE_TYPES) )
            return XmiUtil.FILE_JAVA_PRIMITIVE_TYPES;
        
        else if( path.equals(XmiUtil.LIBRARY_UML2_PRIMITIVE_TYPES) )
            return XmiUtil.FILE_UML2_PRIMITIVE_TYPES;
        
        else
            throw new Exception("!! Cannot resolve resource " + path + " !!");
    }

//    private void loadPrimitiveTypes_(String path) throws Exception
//    {
//        String filename = getResourceOf_( path );
//        XmiModel xModel = xmitools.umlparser.Main.parseModel(filename, true);
//        XmiElement xRoot = xModel.root();
//        
//        Collection items = xRoot.collectChildrenByTagName("ownedMember");
//        for (Iterator iter = items.iterator(); iter.hasNext();)
//        {
//            XmiNode node = (XmiNode) iter.next();
//            if( XmiUtil.getXmiType(node).equals("uml:PrimitiveType") )
//            {
//                String id = XmiUtil.getXmiId( node );
//                nodes.put(id, new XmiUmlPrimitiveType(this, node));
//            }
//        }
//    }
//
//    private String getResourceOf_(String path)
//    {
//        if( path.equals(XmiUtil.LIBRARY_JAVA_PRIMITIVE_TYPES) )
//            return XmiUtil.FILE_JAVA_PRIMITIVE_TYPES_;
//        
//        else if( path.equals(XmiUtil.LIBRARY_UML2_PRIMITIVE_TYPES) )
//            return XmiUtil.FILE_UML2_PRIMITIVE_TYPES_;
//        
//        else {
//            System.err.println("!! Cannot resolve resource " + path + " !!");
//            return null;
//        }
//    }

    public void display() throws Exception
    {
        XmiNodeIterator it = root.iterator();
        while( it.hasNext() )
        {
            XmiNode node = it.next();
            
            // Skip packageImport-s, ownedComment-s...
            if( ! node.tagKind().equals("ownedMember") )
                continue;
            //Element domElem = node.element();
            
            String xmiType = XmiUtil.getXmiType( node );
            //System.err.println("node: " + node.name() + ", " + xmiType );
            
            if( xmiType.equals("uml:Enumeration") )
                XmiUtil.printEnumeration(node);
            
            if( xmiType.equals("uml:Class") )
                printClass(node);
            
        }
        
        System.err.println("ok");
    }

    public void printClass(XmiNode node) throws Exception
    {
        // Get attributes
        Collection attrs = node.collectChildrenByTagName("ownedAttribute");
        
        for (Iterator iter = attrs.iterator(); iter.hasNext();) {
            XmiNode child = (XmiNode) iter.next();
            System.err.println("\t attr: " + child.name());
            
            // Get the "type" attribute
            String typeXmiId = XmiUtil.getTypeAttribute(child);
            if( typeXmiId == null || typeXmiId.equals("") )
            {
                // No "type" attribute, try a <type> child
                XmiNode typeChild = child.getChildByTagName("type");
                if( XmiUtil.isUmlPrimitiveType(typeChild) )
                    printUmlPrimitiveType(typeChild);
            }
            else {
                XmiNode typeNode = XmiUtil.getChildByXmiId(root, typeXmiId);
                System.err.println( XmiUtil.nodeAsString(typeNode) );
            }
        }
    }

    private void printUmlPrimitiveType(XmiNode node) throws Exception
    {
//        String typeName = getUmlPrimitiveTypeNode( node ).name();
//        System.err.println( "primitive type: " + typeName );
    }

    /**
     *  <i>node</i> is a &lt;type&gt; element of form
     *  <br>
     *  <code><type xmi:type="uml:PrimitiveType" href="pathmap://UML2_LIBRARIES/JavaPrimitiveTypes.library.uml2#_TBlMga86EdieaYgxtVWN8Q"/>
     *  </code>
     *  <p> The xmi:type of <i>node</i> must be "uml:PrimitiveType"
     *  
     *  <p> Return the type XmiNode, corresponding to <i>node</i>, defined in the 
     *  uml2 library specified by the <i>href</i> attribute.
     */
//    private XmiNode getUmlPrimitiveTypeNode(XmiNode node) throws Exception
//    {
//        String typeRef = XmiUtil.getAttribute(node, "href"); 
//        
//        if( typeRef == null || typeRef.equals("") )
//            throw new Exception("Attribute 'href' not found for node " + node.name());
//        
//        // Get path
//        String path = null;
//        String id = null;
//        int ind = typeRef.indexOf('#');
//        if( ind != -1 )
//        {
//            path = typeRef.substring(0, ind);
//            id = typeRef.substring( ind+1 );
//        }
//        //System.out.println("\t Found pathmap: " + path + ", id: " + id);
//        
//        if( path != null && path.equals(XmiUtil.LIBRARY_JAVA_PRIMITIVE_TYPES) )
//        {
//            ;//return (XmiNode) getJavaTypeMap().get(id);
//        }
//        
////        if( ! path.equals(XmiUtil.LIBRARY_JAVA_PRIMITIVE_TYPES) && 
////            ! path.equals(XmiUtil.LIBRARY_UML2_PRIMITIVE_TYPES) )
////            throw new Exception("Pathmap " + path + " not known. Expected " +
////                                XmiUtil.LIBRARY_JAVA_PRIMITIVE_TYPES + " or " +
////                                XmiUtil.LIBRARY_UML2_PRIMITIVE_TYPES);
////        
////        return getPrimitiveNode( typeRef ).name();
//        
//        else {
//            System.out.println("Pathmap " + path + " not known. Expected " +
//                                XmiUtil.LIBRARY_JAVA_PRIMITIVE_TYPES);
//            return null;
//        }
//    }

//    private Map getJavaTypeMap() throws Exception
//    {
//        if( javaTypeMap == null )
//        {
//            // Read the java types
//            //System.err.println("reading " + javaTypesFileName);
//            XmiModel javaModel = xmitools.umlparser.Main.parseModel(javaTypesFileName, true);
//            XmiElement javaRoot = javaModel.root();
//            
//            javaTypeMap = new HashMap<String, XmiNode>();
//            
//            Collection items = javaRoot.collectChildrenByTagName("ownedMember");
//            for (Iterator iter = items.iterator(); iter.hasNext();)
//            {
//                XmiNode node = (XmiNode) iter.next();
//                if( XmiUtil.getXmiType(node).equals("uml:PrimitiveType") )
//                {
//                    String id = XmiUtil.getXmiId( node );
//                    javaTypeMap.put(id, node);
//                }
//            }
//        }
//        
//        return javaTypeMap;
//    }

    public XmiUmlStateMachine getUmlStateMachine( String name )
    {
        for( XmiUmlStateMachine usm: umlStateMachines.values() )
        {
            if( usm.name().equals(name) )
                return usm;
        }
        
        return null;
    }
    
    private String getXmiId(XmiNode node)
    {
        return XmiUtil.getXmiId(node);
    }
    
    private String getXmiType(XmiNode node)
    {
        return XmiUtil.getXmiType(node);
    }
}




















