package hu.soft4d.minerva.wsdl;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.wsdl.Message;
import javax.wsdl.Part;
import javax.xml.namespace.QName;

import org.apache.axis.constants.Style;
import org.apache.axis.constants.Use;
import org.apache.axis.wsdl.symbolTable.Element;
import org.apache.axis.wsdl.symbolTable.SymbolTable;
import org.apache.axis.wsdl.symbolTable.TypeEntry;

/**
 * Class representing a WSDL message. Use getParameters() to obtain its parameters
 * under given conditions.  
 * 
 * @author Zsolt Kiss-Gere, 4DSoft Ltd.
 */
public class JMessage
{
    private SymbolTable symTable;
    
    private Message message;
    
    /**
     * {@link Constants#IN_MESSAGE Constants.IN_MESSAGE} or
     * {@link Constants#OUT_MESSAGE Constants.OUT_MESSAGE}
     */
    private int direction;

    
    /**
     * @param message   A javax.wsdl.Message
     * @param direction The message direction: Constants.IN_MESSAGE, Constants.OUT_MESSAGE 
     * @param symTable
     */
    public JMessage( Message message, int direction, SymbolTable symTable )
    {
        this.message = message;
        this.symTable = symTable;
        this.direction = direction;
    }
    
    SymbolTable getSymbolTable()
    {
        return symTable;
    }

    public QName getQName()
    {
        return message.getQName();
    }

    /**
     * Convenience method which returns a list of message parameters if a binding,
     * operation and direction are given.
     * 
     * @param direction  {@link Constants#IN_MESSAGE Constants.IN_MESSAGE} or
     *                   {@link Constants#OUT_MESSAGE Constants.OUT_MESSAGE}
     *                   
     * @return  List of JParameter-s of this message
     */
    public List<JParameter> getParameters( JPortBinding binding, JOperation operation, int direction ) throws Exception
    {
        // port binding style: "rpc" or "document"
        Style style = binding.getBindingStyle();

        Use use = null;
        if( direction == Constants.IN_MESSAGE )
            use = binding.getInputUse( operation );
        else if( direction == Constants.OUT_MESSAGE )
            use = binding.getOutputUse( operation );
        
        return getParameters(use, style);
    }
    
    /**
     * Get the soap:binding style and soap:body for this message.
     * <p>
     * If soap:body use="encoded", then each message part is a parameter and its
     * type attribute specifies the parameter type. The soap:body optional
     * attributes 'parts' and 'namespace' also count.
     * <p>
     * If use="literal", the parts define their type. If a part has an 'element'
     * attribute, this element will appear directly under the body
     * (style=document), or under body | part_name (style=rpc). In the
     * style=document case we assume that there is one part in the message,
     * which contains the parameters.
     * <p>
     * So in the use="literal" case, we also assume a binding style of document
     * and that the message has one part with an 'element' attribute. This will
     * contain the message parameters.
     * <p>
     * <b>Note:</b> it is not possible to get the soap:body element from a wsdl
     * operation or message with the WSDL4J API!
     * 
     * @param use  The "use" attribute of the soap:body tag, defined in an operation
     *             which uses this message.
     *             <br>Can be "encoded" or "literal".
     *             
     * @param style The "style" attribute of the soap:binding tag, defined in a port 
     *              binding.
     *              <br>Can be "rpc" or "document".
     * 
     * @return      A List of JParameter-s of this message.
     */
    public List<JParameter> getParameters( Use use, Style style ) throws Exception
    {
        if( use == Use.LITERAL && style == Style.DOCUMENT )
        {
            // Assume that the message has one part with an 'element' attribute
            Part params = message.getPart( Constants.PARAMETERS );
            if( params == null )
                throw new Exception("Message has no part called 'parameters'");
            
            // <part name="parameters" element="tns:requestQuote"/>
            QName elemName = params.getElementName();
            if( elemName == null )
                throw new Exception("'element' attribute expected for message");
            
            // elemName is the type name of the "parameters" part
            Element typeElement = symTable.getElement( elemName );
            // Get the type entry of this type element
            TypeEntry type = WSDLUtil.getElementType( typeElement );
            
            // This TypeEntry should be a complex type, and its contained elements 
            // are the parameters of the message
            return WSDLUtil.getContainedParameters( type );
        }
        else if( use == Use.ENCODED )
        {
            List<JParameter> params = new ArrayList<JParameter>();
            Map parts = message.getParts();
            for (Iterator iter = parts.keySet().iterator(); iter.hasNext();) {
                String name = (String) iter.next();
                Part part = (Part) parts.get( name );
                QName typeName = part.getTypeName();
                if( typeName == null )
                    throw new Exception("Could not get type for part " + part.getName());
                else
                    params.add( new JParameter(name, typeName) );
            }
            return params;
        }
        else throw new Exception("This use and style combination is not supported yet");
    }
}




















