package hu.soft4d.minerva.wsdl;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.util.logging.Logger;

import javax.xml.namespace.QName;

import org.apache.xerces.xs.XSImplementation;
import org.apache.xerces.xs.XSLoader;
import org.apache.xerces.xs.XSModel;
import org.apache.xerces.xs.XSTypeDefinition;
import org.w3c.dom.DOMConfiguration;
import org.w3c.dom.DOMErrorHandler;
import org.w3c.dom.DOMImplementation;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.bootstrap.DOMImplementationRegistry;
import org.w3c.dom.ls.DOMImplementationLS;
import org.w3c.dom.ls.LSInput;
import org.w3c.dom.ls.LSOutput;
import org.w3c.dom.ls.LSParser;
import org.w3c.dom.ls.LSSerializer;

/**
 * Various utility methods, used in the parsing process.
 * 
 * @author Zsolt Kiss-Gere, 4DSoft Ltd.
 */
public class SchemaHelper
{
    private static Logger log = Logger.getLogger( SchemaHelper.class.getName() );
    
    public static XSModel[] getSchemaModels( String wsdlUri ) throws Exception
    {
        return getSchemaModels( getDOMDocument(wsdlUri), wsdlUri );
    }
    
    public static XSModel[] getSchemaModels( Document document, String wsdlUri ) throws Exception
    {
        NodeList schemaList = document.getElementsByTagNameNS(
                "http://www.w3.org/2001/XMLSchema", "schema");
        //System.out.println("schema count is " + schemaList.getLength());

        XSModel[] models = new XSModel[schemaList.getLength()];
        
        for (int i = 0; i < schemaList.getLength(); i++) {
            Element schemaNode = (Element) schemaList.item(i);
            models[i] = getXSModel(document, schemaNode);
        }

        return models;
    }
    
    public static Document getDOMDocument( String uri ) throws Exception
    {
        DOMImplementationLS implLS = getDOMImplementationLS();

        LSParser parser = implLS.createLSParser( DOMImplementationLS.MODE_SYNCHRONOUS, "http://www.w3.org/2001/XMLSchema");
        DOMConfiguration config = parser.getDomConfig();

        DOMErrorHandler errorHandler = new DOMErrorHandlerImpl();
        config.setParameter("error-handler", errorHandler);

        config.setParameter("validate", Boolean.TRUE);
        config.setParameter("validate-if-schema", Boolean.TRUE);
        config.setParameter("schema-type", "http://www.w3.org/2001/XMLSchema");
        // config.setParameter("schema-location" ,".xsd");
        
        Document document = parser.parseURI( uri );
        // document.normalizeDocument();
        
        return document;
    }
    
    public static DOMImplementationLS getDOMImplementationLS() throws Exception
    {
        System.setProperty(DOMImplementationRegistry.PROPERTY, "org.apache.xerces.dom.DOMImplementationSourceImpl");
        DOMImplementationRegistry registry = DOMImplementationRegistry.newInstance();

        DOMImplementation domImpl = registry.getDOMImplementation("LS 3.0");
        DOMImplementationLS implLS = (DOMImplementationLS) domImpl; 
        
        return implLS;
    }
    
    private static XSModel getXSModel(Document document, Element schemaNode) throws Exception
    {
        //System.out.println("entered getXSModel");
        
        // Apply all root namespace attributes to the schema, if they are not there already 
        Element root = document.getDocumentElement();
        NamedNodeMap attrmap = root.getAttributes();
        
        for( int i=0; i < attrmap.getLength(); i++ )
        {
            Node attrnode = attrmap.item(i);
            String name = attrnode.getNodeName();
            String value = attrnode.getNodeValue();
            
            if( name.startsWith("xmlns") && !value.equals("http://www.w3.org/2001/XMLSchema") )
                if( schemaNode.getAttributeNode(name) == null )
                {
                    //System.out.println("Setting attr " + name + " to " + value);
                    schemaNode.setAttribute( name, value );
                }
                else
                    ; //System.out.println("Attr " + name + " not overridden");
            
            // The following does NOT work:
            // schemaNode.setAttributeNS(null, "xmlns:soapenc", "http://schemas.xmlsoap.org/soap/encoding/");
        }
        
        DOMImplementationLS implLS = getDOMImplementationLS();
        
        LSOutput output = implLS.createLSOutput();
        output.setByteStream( new ByteArrayOutputStream() );
        output.setEncoding("UTF-8");
        
        LSSerializer dom3Writer = implLS.createLSSerializer();
        dom3Writer.write( schemaNode, output );
        
        //String domstr = ((ByteArrayOutputStream) output.getByteStream()).toString("UTF-8");
        //System.out.println( domstr );
        
        byte[] buf = ((ByteArrayOutputStream) output.getByteStream()).toByteArray();
        LSInput lsinput = implLS.createLSInput();
        lsinput.setByteStream(new ByteArrayInputStream(buf));
        lsinput.setEncoding("UTF-8");
        
        XSLoader xsLoader = getXSLoader();
        XSModel model = xsLoader.load( lsinput );
        return model;
    }
    
    private static XSLoader getXSLoader() throws Exception
    {
        System.setProperty(DOMImplementationRegistry.PROPERTY,
                "org.apache.xerces.dom.DOMXSImplementationSourceImpl");

        DOMImplementationRegistry registry = DOMImplementationRegistry.newInstance();
        XSImplementation impl = (XSImplementation) registry.getDOMImplementation("XS-Loader");
        
        XSLoader schemaLoader = impl.createXSLoader(null);
        DOMConfiguration config = schemaLoader.getConfig();

        // create Error Handler
        DOMErrorHandler errorHandler = new DOMErrorHandlerImpl();
        // set error handler
        config.setParameter("error-handler", errorHandler);
        // set validation feature
        config.setParameter("validate", Boolean.TRUE);
        
        return schemaLoader;
    }

    /**
     * Looks for the type given by qname in the schema models and returns the first one found,
     * or null. 
     */
    public static XSTypeDefinition getTypeDefinition( XSModel[] schemas, QName qname )
    {
        for (int i = 0; i < schemas.length; i++) 
        {
            XSTypeDefinition type = schemas[i].getTypeDefinition( qname.getLocalPart(), 
                                                                  qname.getNamespaceURI());
            if( type != null )  return type;
        }
        log.warning("Type " + qname + " was not found in the schema models");
        return null;
    }
}



















