package hu.soft4d.minerva.wsdl;
/**
 * Copyright (C) 2007 4D Soft Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * See the LICENSE file for the full license
 */


import java.util.List;
import java.util.Map;

import javax.wsdl.OperationType;
import javax.xml.namespace.QName;

import org.apache.axis.constants.Style;
import org.apache.axis.constants.Use;
import org.apache.axis.wsdl.symbolTable.DefinedType;
import org.apache.xerces.xs.StringList;
import org.apache.xerces.xs.XSModel;
import org.apache.xerces.xs.XSSimpleTypeDefinition;
import org.apache.xerces.xs.XSTypeDefinition;


/**
 * Command line tool to parse a WSDL and show its structure.
 * <p> For testing the wsdl package.
 * 
 * @author Zsolt Kiss-Gere, 4DSoft Ltd.
 */
public class ShowWSDL
{
    private WSDLModel wsdl;
    
    public static void main(String[] args)
    {
        if (args.length < 1) {
            System.err.println("usage: java hu.soft4d.minerva.wsdl.ShowWSDL <wsdl-uri>");
            System.exit(2);
        }
        String uri = args[0];

        ShowWSDL wsh = new ShowWSDL();
        try {
            wsh.showWSDL( uri );
        } 
        catch (Exception e) {
            e.printStackTrace();
        }
    }

    public void showWSDL( String uri ) throws Exception
    {
        wsdl = WSDLModelFactory.newWSDLModel( uri );
        
        // Show all services, with their enclosed components
        Map<QName, JService> services = wsdl.getServices();
        if( services.size() == 0 ) {
            System.out.println("Services not found");
            return;
        }
        else for (JService service: services.values()) {
            showService( service );
        }
        
        // Show all defined types
        Map<QName, DefinedType> defTypes = wsdl.getAllDefinedTypes();
        showTypes( defTypes );
    }

    private void showTypes( Map<QName, DefinedType> defTypes ) throws Exception
    {
        for (Map.Entry<QName, DefinedType> entry : defTypes.entrySet())
        {
            QName qname = entry.getKey();
            System.out.println("Type: " + qname.getLocalPart() );
            
            DefinedType typeEntry = defTypes.get( qname );
            
            if( ! typeEntry.isBaseType() && ! typeEntry.isSimpleType() ) 
            {
                // Complex type
                List<JParameter> params = WSDLUtil.getContainedParameters( typeEntry );
                showParameters( params );
            }
            else {
                showType( typeEntry.getQName() );
            }
        }
    }

    private void showType( QName qname ) throws Exception 
    {
        XSModel[] schemas = wsdl.getSchemaModels();
        XSTypeDefinition type = SchemaHelper.getTypeDefinition( schemas, qname ); 
        
        if( type != null && type.getTypeCategory() == XSTypeDefinition.SIMPLE_TYPE )
        {
            XSSimpleTypeDefinition simpleType = (XSSimpleTypeDefinition) type;
            StringList slist = simpleType.getLexicalEnumeration();
            if( slist != null && slist.getLength() > 0 )
            {
                System.out.println("\t\t\tENUMERATION:");
                for (int j = 0; j < slist.getLength(); j++) {
                    System.out.println("\t\t\t\t" + slist.item(j));
                }
            }
        }
	}

	private void showService( JService service ) throws Exception
    {
        System.out.println("Service: " + service.getQName().getLocalPart() );
        
        Map<String, JPort> ports = service.getPorts();
        
        for (JPort port: ports.values()) {
            showPort( port );
        }
    }

    private void showPort( JPort port ) throws Exception
    {
        JPortBinding binding = port.getBinding();
        // We are interested only in SOAP bindings
        if( ! binding.isSOAPBinding() )
            return;
        
        JInterface intface = binding.getInterface();
        System.out.println("  Port: " + port.getName() + " (interface " + intface.getLocalName() + ")");
        
        List<JOperation> operations = intface.getOperations();
        for (JOperation operation : operations) {
            showOperation(binding, operation);
        }
    }

    private void showOperation( JPortBinding binding, JOperation operation ) throws Exception
    {
        // operation type: REQUEST_RESPONSE, etc.
        OperationType optype = operation.getType();
        System.out.println("     " + operation.getName() + ":");
        
        // port binding style: "rpc" or "document"
        Style style = binding.getBindingStyle();
        
        if (optype != OperationType.NOTIFICATION) 
        {
            JMessage inmsg = operation.getInputMessage();
            // soap:body use: "literal" or "encoded"
            Use use = binding.getInputUse( operation );
            System.out.println("         in msg: " + inmsg.getQName().getLocalPart());
            List<JParameter> params = inmsg.getParameters(use, style);
            showParameters( params );
        }
        if (optype != OperationType.ONE_WAY) {
            JMessage outmsg = operation.getOutputMessage();
            Use use = binding.getOutputUse( operation );
            System.out.println("        out msg: " + outmsg.getQName().getLocalPart());
            List<JParameter> params = outmsg.getParameters(use, style);
            showParameters( params );
        }
    }

    private void showParameters( List<JParameter> params )
    {
        for (JParameter par : params) {
            System.out.println("\t\t\t" + par.getName() + ": " + par.getTypeName() );
        }
    }
}
















