package hu.soft4d.minerva.wsdl;

import hu.soft4d.minerva.wsdl.ssm.SSMHelper;

import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Vector;
import java.util.logging.Logger;

import javax.wsdl.Service;
import javax.xml.namespace.QName;

import org.apache.axis.wsdl.gen.Parser;
import org.apache.axis.wsdl.symbolTable.DefinedType;
import org.apache.axis.wsdl.symbolTable.SymbolTable;
import org.apache.axis.wsdl.symbolTable.TypeEntry;
import org.apache.xerces.xs.StringList;
import org.apache.xerces.xs.XSModel;
import org.apache.xerces.xs.XSSimpleTypeDefinition;
import org.apache.xerces.xs.XSTypeDefinition;
import org.w3c.dom.Document;

import stsimulator.types.ComplexType;
import stsimulator.types.Enumeration;
import stsimulator.types.Type;


/**
 * An implementation of {@link WSDLModel WSDLModel}. Represents the structure of a WSDL file.
 * 
 * @author Zsolt Kiss-Gere, 4DSoft Ltd.
 */
public class WSDLModelImpl implements WSDLModel
{
    private static Logger log = Logger.getLogger( WSDLModelImpl.class.getName() );
    
    private String wsdlUri;
    private SymbolTable symbolTable;
    private Document document;
    private XSModel[] schemaList;
    private Map<QName, Type> ssmTypes = new HashMap<QName, Type>();
    
    /**
     * Parses the WSDL stream defined by wsdlURI and builds a memory object from it.
     */
    WSDLModelImpl( String wsdlURI ) throws Exception
    {
        Parser wsdlParser = new Parser();
        log.info("Reading WSDL document from '" + wsdlURI + "' ...");
        wsdlParser.run( wsdlURI );
        this.wsdlUri = wsdlURI;
        symbolTable = wsdlParser.getSymbolTable();
        //symbolTable.dump(System.out);
    }

    public SymbolTable getSymbolTable()
    {
        return symbolTable;
    }

    public JService getService(String serviceName) throws Exception
    {
        QName sname = new QName(getTargetNamespace(), serviceName);
        Service service = getSymbolTable().getDefinition().getService(sname);
        if( service == null )
            throw new Exception("No service named " + serviceName + " was found");
        
        return new JService(this, service);
    }

    public Map<QName, JService> getServices()
    {
        Map services = getSymbolTable().getDefinition().getAllServices();
        Map<QName, JService> map = new HashMap<QName, JService>();
        
        for (Iterator iter = services.keySet().iterator(); iter.hasNext();) {
            QName sname = (QName) iter.next();
            Service service = (Service) services.get( sname );
            map.put(sname, new JService(this, service));
        }
        return map;
    }

    public Map<QName, DefinedType> getAllDefinedTypes()
    {
        Map<QName, DefinedType> defs = new HashMap<QName, DefinedType>();
        HashMap map = symbolTable.getHashMap();
        
        for (Iterator iter = map.keySet().iterator(); iter.hasNext();) {
            QName name = (QName) iter.next();
            Vector list = (Vector) map.get(name);
            if( list.size() == 0 ) 
                continue;
            // The size should be 1 for defined types
            Object entry = list.get(0);
            if( ! (entry instanceof DefinedType) ) 
                continue;
            
            DefinedType deftype = (DefinedType) entry;
            
            // If entry.localName starts with ">", it is an embedded type in an element,
            // (RefType for a DefinedElement) so we don't need it
            if( deftype.getQName().getLocalPart().startsWith(">") )
                continue;
            
            defs.put( name, deftype );
        }
        return defs;
    }

    public XSModel[] getSchemaModels() throws Exception
    {
        if( schemaList == null )
            schemaList = SchemaHelper.getSchemaModels( getDOMDocument(), wsdlUri ); 
        
        return schemaList;
    }

    private Document getDOMDocument() throws Exception
    {
        if( document == null )
            document = SchemaHelper.getDOMDocument( wsdlUri );
        
        return document;
    }

    public String getTargetNamespace() throws Exception
    {
        return getSymbolTable().getDefinition().getTargetNamespace();
    }

    public Type getCreateSSMType(QName typeQName) throws Exception
    {
        //return getSSMType( getTypeEntry(typeQName) );
        TypeEntry te = getTypeEntry(typeQName);
        Type t = getCreateSSMType( te );
        return t;
    }

    private TypeEntry getTypeEntry( QName typeQName ) throws Exception
    {
        TypeEntry tEntry = symbolTable.getType( typeQName );
        if( tEntry != null )
            return tEntry;
        else
            throw new Exception("Could not get type entry for " + typeQName);
    }

    /**
     * Return an SSM Type object corresponding to TypeEntry, using an internal Map for cache.
     *   
     * <p> If the type already exists in the map (key=typeEntry.qname), just return it.
     * <p> Else construct a new Simple- or ComplexType, put it in the map and return it.
     * 
     * <p> Traverses recursively the tree of embedded parameter types, constructing
     *     new types as needed.
     * 
     * @see hu.soft4d.minerva.wsdl.WSDLModel#getCreateSSMType(org.apache.axis.wsdl.symbolTable.TypeEntry)
     */
    public Type getCreateSSMType(TypeEntry typeEntry) throws Exception
    {
        return getCreateSSMType2(typeEntry, ssmTypes);
    }

    /**
     * Get the SSM type corresponding to type entry, using a map, given as the 
     * second parameter for cache. This method may call itself recursively, if the type
     * is a complex one embedding other types. Each calculated SSM type is stored in the map, 
     * which is then passed down the tree. If a type is contained in the map, it is returned
     * and no calculations are made.
     */
    private Type getCreateSSMType2(TypeEntry typeEntry, Map<QName, Type> ssmMap) throws Exception
    {
        QName qname = typeEntry.getQName();
        
        if( ssmMap.containsKey(qname) )
            return ssmMap.get(qname);
        
        if( typeEntry.isBaseType() || typeEntry.isSimpleType() )
        {
            Type ssmType = getSSMSimpleType(typeEntry);
            ssmMap.put(qname, ssmType);
            return ssmType;
        }
        
        // typeEntry is a complex type
        ComplexType complex = new ComplexType( qname );
        ssmMap.put(qname, complex);
        
        // Get the JParameterList of typeEntry
        List<JParameter> jparList = WSDLUtil.getContainedParameters( typeEntry );
        Object[][] parArray = new Object[jparList.size()][2];
        int i = 0;
        for (Iterator iter = jparList.iterator(); iter.hasNext(); i++) 
        {
            JParameter par = (JParameter) iter.next();
            String name = par.getName();
            QName typeQName = par.getTypeQName();
            TypeEntry tEntry = getTypeEntry( typeQName );
            // Now call recursively
            Type parSSMType = getCreateSSMType2(tEntry, ssmMap);
            // Ok, now we are ready to construct an SSM parameter with (name, parSSMType)
            parArray[i][0] = name;
            parArray[i][1] = parSSMType;
        }
        complex.setFields( parArray );
        
        return complex;
    }
    
    /**
     * Get the SSM Type corresponding to this type entry, which should be a
     * simple type. The SSM type should result from a mapping table.
     * 
     * @param typeEntry  Should be a simple type.
     * @return           The SSM Type object corresponding to this type entry.
     */
    private Type getSSMSimpleType(TypeEntry typeEntry) throws Exception
    {
        QName qname = typeEntry.getQName();
        String nspace = typeEntry.getQName().getNamespaceURI();
        if( nspace.equals(Constants.XML_SCHEMA) )
        {
            // XML Schema simple type
            String name = typeEntry.getQName().getLocalPart();
            String className = SSMHelper.getSSMClassName( name );
            
            return (Type) Class.forName(className).newInstance();
        }
        else {
            // Try an enumeration type
            XSModel[] schemas = getSchemaModels();
            XSTypeDefinition typedef = SchemaHelper.getTypeDefinition( schemas, qname );
            
            if( typedef == null )
                throw new Exception("Type " + qname + " was not found");
            
            if( typedef.getTypeCategory() == XSTypeDefinition.SIMPLE_TYPE )
            {
                XSSimpleTypeDefinition simpleType = (XSSimpleTypeDefinition) typedef;
                StringList slist = simpleType.getLexicalEnumeration();
                if( slist != null && slist.getLength() > 0 )
                {
                    return getEnumeration(qname, slist);
                }
                else
                    throw new Exception("Only enumeration or an XML Schema simple type expected." +
                                        "\n Got " + qname );
            }
            else
                throw new Exception("Simple type expected");  // should never happen
        }
    }

    private Type getEnumeration(QName qname, StringList slist)
    {
        int size = slist.getLength();
        String[] items = new String[size];
        
        for (int i = 0; i < size; i++) {
            items[i] = slist.item(i);
        }
        
        return new Enumeration(qname, items);
    }

    /**
     * @see hu.soft4d.minerva.wsdl.WSDLModel#getSSMTypes()
     */
    public Map<QName, Type> getSSMTypes()
    {
        return ssmTypes;
    }

}



















