/**
 * Copyright (C) 2007 4D Soft Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * See the LICENSE file for the full license
 */

package hu.soft4d.minerva.wsdl.ssm;

import hu.soft4d.minerva.wsdl.Constants;
import hu.soft4d.minerva.wsdl.JOperation;

import stsimulator.sts.*;
import stsimulator.types.*;

import java.util.Iterator;
import java.util.Map;
import javax.wsdl.OperationType;
import javax.xml.namespace.QName;

/**
 * Helper class containing different static utilities for conversion, etc.
 * 
 * @author Zsolt Kiss Gere, 4D Soft Ltd.
 */
public class SSMHelper
{
    public static Operation getSSMOperation( JOperation op )
    {
        OperationType opType = op.getType();
        OperationKind kind = null;
        
        if( opType == OperationType.REQUEST_RESPONSE )
            kind = OperationKind.REQUESTRESPONSE;
        
        else if( opType == OperationType.ONE_WAY )
            kind = OperationKind.ONEWAY;
        
        else if( opType == OperationType.NOTIFICATION )
            kind = OperationKind.NOTIFICATION;
        
        else if( opType == OperationType.SOLICIT_RESPONSE )
            kind = OperationKind.SOLICITRESPONSE;
        
        return new Operation(op.getName(), kind);
    }
    
    public static int getDirection( String dir )
    {
        if( dir.equals(Constants.INPUT) )
            return Constants.IN_MESSAGE;
        
        else if( dir.equals(Constants.OUTPUT) )
            return Constants.OUT_MESSAGE;
        
        else return -2;  // should not happen
    }
    
    public static MessageKind getMessageKind( String dir ) throws Exception
    {
        if( dir.equals(Constants.INPUT) )
            return MessageKind.INPUT;
        
        else if( dir.equals(Constants.OUTPUT) )
            return MessageKind.OUTPUT;
        
        else if( dir.equals(Constants.UNOBSERVABLE) )
            return MessageKind.UNOBSERVABLE;
        
        else
            throw new Exception("Unknown direction " + dir );
    }
    
    /**
     * This is a mapping from an XML Schema simple type to an SSM type name.
     * 
     * @param name  The local name of an XML Schema simple type. Ex: "int", "boolean", etc.
     * @return  The qualified class name of the SSM Type which corresponds to the simple type.
     */
    public static String getSSMClassName(String name) throws Exception
    {
        // A mapping table contained in a property file is preferred
        
        String p = Constants.SSM_TYPES_PACKAGE + ".";
        
        if(name.equals("string") || name.equals("String")) return p + "ST_String";
        
        else if(name.equals("int") || name.equals("Integer"))  return p + "ST_Int";
        
        else if(name.equals("positiveInteger")) return p + "ST_PosInt";
        
        else if(name.equals("boolean") || name.equals("Boolean")) return p + "ST_Boolean";
        
        else if(name.equals("double")) return p + "ST_PseudoPosDouble";  // instead of ST_Double
        
        else throw new Exception("Sorry, type " + name + " is not mapped yet");
    }
    
    public static Type getSSMSimpleType( String name ) throws Exception
    {
        String className = getSSMClassName( name );
        
        return (Type) Class.forName(className).newInstance();
    }
    
    public static void addInteractionVariables( SymbolicTransitionSystem ssm, 
                                                Map<String, InteractionVariable> intvarMap)
    {
        for( InteractionVariable var: intvarMap.values() )
            ssm.addVariable( var );
    }

    public static void addTypes(SymbolicTransitionSystem ssm, Map<QName, Type> types)
    {
        for (Iterator iter = types.values().iterator(); iter.hasNext();) 
        {
            Type type = (Type) iter.next();
            ssm.addType( type );
        }
    }

    public static void addMessages(SymbolicTransitionSystem ssm, Map<String, Message> messageMap)
    {
        for (Iterator iter = messageMap.values().iterator(); iter.hasNext();) 
        {
            Message msg = (Message) iter.next();
            ssm.addMessage( msg );
        }
    }

    public static void addOperations(SymbolicTransitionSystem ssm, Map<String, Operation> operationMap)
    {
        for (Iterator iter = operationMap.values().iterator(); iter.hasNext();) 
        {
            Operation op = (Operation) iter.next();
            ssm.addOperation( op );
        }
    }
}
















