/**
 * Copyright (C) 2007 4D Soft Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * See the LICENSE file for the full license
 */

package hu.soft4d.minerva.wsdl.ssm;

import hu.soft4d.util.JDomHelper;

import java.net.URL;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.logging.Logger;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.input.SAXBuilder;


/**
 * This class takes an SSM schema definition URL, parses the stream and
 * builds a memory object model of the data found in the schema instance.
 * This can then be queried.
 * 
 * @author Zsolt Kiss Gere, 4D Soft Ltd.
 */
public class SSMSchemaModel
{
    private static Logger log = Logger.getLogger(SSMSchemaModel.class.getName());
    
    private Document doc;
    private Element root;
    private String serviceName;
    private String portName;
    
    /**
     * Parse the XML document identified by schemaUrl. Currently uses JDom.
     * @param serviceName 
     */
    public SSMSchemaModel( URL schemaUrl, String serviceName, String portName ) throws Exception
    {
        this.serviceName = serviceName;
        this.portName = portName;
        SAXBuilder builder = new SAXBuilder();
        log.info("Parsing " + schemaUrl + "...");
        doc = builder.build( schemaUrl );
        root = doc.getRootElement();
    }

    public String getWSDLUrlString() throws Exception
    {
        return getFirstLevelElementValue("wsdlURI");
    }

    /*
     * The service name is given now as a parameter; it is no longer in the ssm xml.
     */
    public String getServiceName() throws Exception
    {
        return serviceName;
    }
    
    
    public String getInitialState() throws Exception
    {
        return getFirstLevelElementValue("initialState");
    }

    public List<LocationVarEntry> getLocVarEntries() throws Exception
    {
        List<LocationVarEntry> list = new ArrayList<LocationVarEntry>();
        Element variables = JDomHelper.getChild(root, "variables");
        List children = variables.getChildren();
        // These should be "var" elements
        for (Iterator iter = children.iterator(); iter.hasNext();) 
        {
            Element child = (Element) iter.next();
            if(! child.getName().equals("var") )
                throw new Exception("Only var elements allowed in variables");
            
            LocationVarEntry entry = getLocVarEntry( child );
            list.add( entry );
        }
        return list;
    }
    
    public List<SwitchEntry> getSwitchEntries() throws Exception
    {
        String serviceName = getServiceName();
        List<SwitchEntry> list = new ArrayList<SwitchEntry>();
        Element switches = JDomHelper.getChild(root, "switches");
        List children = switches.getChildren();
        // These should be "switch" elements
        for (Iterator iter = children.iterator(); iter.hasNext();) 
        {
            Element child = (Element) iter.next();
            if(! child.getName().equals("switch") )
                throw new Exception("Only switch elements allowed in switches");
            
            SwitchEntry entry = getSwitchEntry(serviceName, child);
            list.add( entry );
        }
        return list;
    }

    private LocationVarEntry getLocVarEntry(Element locvar) throws Exception
    {
        Element nameEl = JDomHelper.getChild(locvar, "name");
        Element typeEl = JDomHelper.getChild(locvar, "type");
        
        return new LocationVarEntry(locvar, nameEl.getTextTrim(), typeEl.getTextTrim());
    }

    private SwitchEntry getSwitchEntry( String serviceName, Element switchEl ) throws Exception
    {
        Element fromEl = JDomHelper.getChild(switchEl, "from");
        Element targetEl = JDomHelper.getChild(switchEl, "to");
        Element operationNameEl = JDomHelper.getChild(switchEl, "operationName");
        Element directionEl = JDomHelper.getChild(switchEl, "kind");
        Element restrictionEl = JDomHelper.getChild(switchEl, "restriction");
        Element updateEl = JDomHelper.getChild(switchEl, "update");
        
        return new SwitchEntry( serviceName, fromEl.getTextTrim(), targetEl.getTextTrim(),  
                                operationNameEl.getTextTrim(), directionEl.getTextTrim(), 
                                restrictionEl.getTextTrim(), 
                                updateEl.getTextTrim() );
    }
    
    private String getFirstLevelElementValue( String name ) throws Exception
    {
        return JDomHelper.getChild(root, name).getTextTrim();
    }
}
