/**
 * Copyright (C) 2007 4D Soft Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * See the LICENSE file for the full license
 */

package hu.soft4d.minerva.wsdl.ssm;

import hu.soft4d.minerva.ssm.SSMFactory;
import hu.soft4d.minerva.wsdl.Constants;
import hu.soft4d.minerva.wsdl.JInterface;
import hu.soft4d.minerva.wsdl.JMessage;
import hu.soft4d.minerva.wsdl.JOperation;
import hu.soft4d.minerva.wsdl.JParameter;
import hu.soft4d.minerva.wsdl.JPort;
import hu.soft4d.minerva.wsdl.JPortBinding;
import hu.soft4d.minerva.wsdl.JService;
import hu.soft4d.minerva.wsdl.WSDLModel;
import hu.soft4d.minerva.wsdl.WSDLModelFactory;
import hu.soft4d.util.Sequence;

import stsimulator.sts.*;
import stsimulator.types.*;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Logger;

import javax.xml.namespace.QName;

import org.jdom.Element;
import org.jdom.Namespace;


/**
 * Factory class for generating a SymbolicTransitionSystem, if an SSM schema
 * definition URL is given.
 * 
 * <p>The generated ssm is not checked for consistency, neither it has
 * switch parsers attached.
 * 
 * @author Zsolt Kiss Gere, 4D Soft Ltd.
 */
public class WSDLSSMFactory extends SSMFactory
{
    private static Logger log = Logger.getLogger("WSDLSSMFactory");
    
    private WSDLModel wsdl;
    private JService jservice;
    private JPort jport;
    private SSMSchemaModel ssmModel;
    
    public WSDLSSMFactory(String wsdlUri, String schemaInstanceUri, String serviceName, String portName) throws Exception
    {
        this.wsdlUri = wsdlUri;
        this.serviceName = serviceName;
        this.portName = portName;
        
        // Reset counters
        Sequence.reset();
        
        URL schemaURL = new URL(schemaInstanceUri);
        ssmModel = new SSMSchemaModel( schemaURL, serviceName, portName );
        initialize();
    }

    /**
     * Fill the containers.
     */
    private void initialize() throws Exception
    {
        namespace = getWSDL().getTargetNamespace();
        log.info("Got targetNamespace '" + namespace + "'");
        
        // The service and port
        jservice = getWSDL().getService( serviceName );
        jport = jservice.getPort( portName );
        
        port = new Port( portName );
        service = new Service( serviceName );
        service.addPort( port );
        
        // The location variables are the global variables
        List<LocationVarEntry> locvarEntries = ssmModel.getLocVarEntries();
        for( LocationVarEntry entry: locvarEntries )
        {
            // The following call fills the locvarMap
            LocationVariable locvar = getCreateLocationVar( entry );
        }
        
        // Create the switches and in the meantime populate the location map
        List<SwitchEntry> switchEntries = ssmModel.getSwitchEntries();
        for (SwitchEntry entry: switchEntries )
        {
            // Switches
            Switch sswitch = createSwitch( entry );
            switchList.add( sswitch );
            
            // Locations
            String start = entry.startLocation;
            String end = entry.targetLocation;
            
            // Store new Location objects if not already there
            if(!locationMap.containsKey(start))
                locationMap.put(start, new Location(start));
            
            if(!locationMap.containsKey(end))
                locationMap.put(end, new Location(end));
            
            // Insert switch - location associations
            Location startLoc = locationMap.get(start);
            startLoc.addOutgoingSwitch( sswitch );
            
            Location endLoc = locationMap.get(end);
            sswitch.setTarget( endLoc );
        }
        
        typeMap = wsdl.getSSMTypes();
        
        // setCurrentLocation
        initialState = ssmModel.getInitialState();
    }
    
    /**
     * Location variables are the global variables
     */
    private LocationVariable getCreateLocationVar(LocationVarEntry entry) throws Exception
    {
        String name = entry.name;
        // type is supposed to be of form "prefix:name"
        String type = entry.type;
        // The "var" element
        Element parent = entry.parent;
        
        // Get the type namespace
        String prefix;
        String typename;
        String[] sarr = type.split(":",2);
        // sarr.length can be 1 or 2
        if( sarr.length > 1 ) {
            prefix = sarr[0];
            typename = sarr[1];
        }
        else {
            prefix = "";
            typename = type;
        }
        //log.info("Got type: {" + prefix + "}" + typename);
        
        // The type namespace
        Namespace typeNsp = null;
        if( prefix.equals("") )
            typeNsp = parent.getNamespace();  // the default namespace
        else
            typeNsp = parent.getNamespace( prefix );
        
        if( typeNsp == null )
            throw new Exception("Could not resolve namespace for prefix '" + prefix + "'");
        //log.info("The type namespace is " + typeNsp );
        
        QName qname = new QName(typeNsp.getURI(), typename);
        // This call fills ssmTypes in object wsdl
        Type t = wsdl.getCreateSSMType( qname );
        
        // The following call to super fills the locvarMap
        LocationVariable var = getCreateLocationVariable(name, t);
        return var;
    }

    private Switch createSwitch( SwitchEntry entry ) throws Exception
    {
        // First, we need a Message. We get operation, kind from the switch entry
        Message ssmMsg = null;
        
        if( entry.kind.equals(Constants.UNOBSERVABLE) )
        {
            ssmMsg = Message.generateUnobservableMessageForPort(port);
            // Can one unobservable message be used more times in operations?
            messageMap.put("", ssmMsg);
        }
        else {
            // An (operation, kind) identifies a(n ssm) message
            // which could have been already stored
            String msgKey = entry.operationName + "@" + entry.kind;
            
            ssmMsg = messageMap.get(msgKey);
            if( ssmMsg == null )
            {
                ssmMsg = createSSMMessage( entry );
                messageMap.put(msgKey, ssmMsg);
            }
        }
        
        String restriction = entry.restriction;
        String update = entry.updateMapping;
        
        Switch sswitch = new Switch( ssmMsg, restriction, update );
        return sswitch;
    }
    
    /**
     *  Create an SSM message corresponding to the given switch entry.
     */
    private Message createSSMMessage( SwitchEntry entry ) throws Exception
    {
        JPortBinding binding = jport.getBinding();
        // We are interested only in SOAP bindings
        if( ! binding.isSOAPBinding() )
            throw new Exception("Sorry, we are interested only in SOAP bindings");
        
        JInterface intface = binding.getInterface();
        JOperation op = intface.getOperation( entry.operationName );
        
        JMessage msg = null;
        if( entry.kind.equals(Constants.INPUT) )
            msg = op.getInputMessage();
        else if( entry.kind.equals(Constants.OUTPUT) )
            msg = op.getOutputMessage();
        else
            throw new Exception("Kind '" + entry.kind + "' not known. " +
                    "\n Expected " + Constants.INPUT + ", " + Constants.OUTPUT +
                    " or " + Constants.UNOBSERVABLE);
        
        String msgName = msg.getQName().getLocalPart();
        
        int dir = SSMHelper.getDirection( entry.kind );
        MessageKind kind = SSMHelper.getMessageKind( entry.kind );
        
        List<JParameter> msgParams = msg.getParameters(binding, op, dir);
        
        // The SSM parameter list
        ArrayList<InteractionVariable> paramList = new ArrayList<InteractionVariable>();
        
        // For each parameter, create an interaction variable and save it
        for( JParameter jpar: msgParams )
        {
            String name = jpar.getName();
            QName typename = jpar.getTypeQName();
            Type type = getWSDL().getCreateSSMType( typename );
            
            InteractionVariable var = getCreateInteractionVariable(name, type);
            paramList.add( var );
        }
        
        Operation oper = getCreateSSMOperation(op.getName(), op.getOperationKind());

        Message ssmMsg = new Message( msgName, kind, oper, port, paramList );
        return ssmMsg;
    }

    private WSDLModel getWSDL() throws Exception
    {
        if( wsdl == null )
            wsdl = WSDLModelFactory.newWSDLModel( wsdlUri );
        
        return wsdl;
    }
    
}













