/**
 * Copyright (C) 2007 4D Soft Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * See the LICENSE file for the full license
 */

package hu.soft4d.minerva.xmi;

import java.util.Collection;
import java.util.Iterator;
import java.util.logging.Logger;

import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;

import xmitools.xmiparser.XmiModel;
import xmitools.xmiparser.XmiNode;
import xmitools.xmiparser.XmiNodeIterator;

/**
 * @author Zsolt Kiss Gere, 4D Soft Ltd.
 */
public class XmiUtil
{
    private static Logger log = Logger.getLogger("XmiUtil"); 
    
    public static final String LIBRARY_JAVA_PRIMITIVE_TYPES = "pathmap://UML2_LIBRARIES/JavaPrimitiveTypes.library.uml2";
    public static final String LIBRARY_UML2_PRIMITIVE_TYPES = "pathmap://UML2_LIBRARIES/UML2PrimitiveTypes.library.uml2";
    
    public static final String FILE_JAVA_PRIMITIVE_TYPES = "JavaPrimitiveTypes.library.uml2";
    public static final String FILE_UML2_PRIMITIVE_TYPES = "UML2PrimitiveTypes.library.uml2";

//    public static final String FILE_JAVA_PRIMITIVE_TYPES_ = "MagicDraw/libraries/JavaPrimitiveTypes.library.uml2";
//    public static final String FILE_UML2_PRIMITIVE_TYPES_ = "MagicDraw/libraries/UML2PrimitiveTypes.library.uml2";

    public static void printEnumeration(XmiNode node)
    {
        Collection items = node.collectChildrenByTagName("ownedLiteral");
        
        for (Iterator iter = items.iterator(); iter.hasNext();) {
            XmiNode child = (XmiNode) iter.next();
            System.err.println("\t item: " + child.name());
        }
    }
    
    public static void printModel(XmiModel model)
    {
        //model.prettyPrint(new PrintStream(new FileOutputStream(reportFileName)));
        model.prettyPrint( System.out );
    }

    public static String getXmiType(XmiNode node)
    {
        return getAttribute(node, "xmi:type");
    }

    public static String getXmiId(XmiNode node)
    {
        return getAttribute(node, "xmi:id");
    }

    public static String getTypeAttribute(XmiNode node)
    {
        return getAttribute(node, "type");
    }
    
    public static String getAttribute(XmiNode node, String attr)
    {
        Element e = node.element();
        String nodename = e.getNodeName();
        String a = e.getAttribute(attr);
        return node.element().getAttribute( attr );
    }
    
    public static String nodeAsString(XmiNode node)
    {
        return "node: " + node.name() + ", " + node.pathName() 
                + ", " + node.tagName();
    }

    /**
     *  Searches the first child with the given xmi:id. Return null if
     *  none is found. The given node is not checked. 
     *  
     *  @return the first child with the given xmi:id, or null xmiId is null 
     *          or none is found
     */
    public static XmiNode getChildByXmiId(XmiNode node, String xmiId)
    {
        if( xmiId == null )  return null;
        
        for (XmiNodeIterator iter = node.iterator(); iter.hasNext();) 
        {
            XmiNode child = iter.next();
            
            String id = getXmiId(child);
            if (id != null && id.equals(xmiId))
                return child;
    
            // Check the children tree below by a recursive call
            XmiNode nested = getChildByXmiId(child, xmiId);
            if (nested != null)
                return nested;
        }
    
        // Return null if neither this node, neither any child node below has
        // the xmiId given
        return null;
    }

    protected static NamedNodeMap printDomAttributes(Element domElem)
    {
        NamedNodeMap attributes = domElem.getAttributes();
        for (int i = 0; i < attributes.getLength(); i++) {
            Node attr = attributes.item(i);
            System.err.println("Attr: " + attr.getNodeName());
        }
        return attributes;
    }
    
    /**
     * Resembles UmlNode.getTag()
     */
    public static boolean isOfStereotype( XmiNode node, String profile, String stereotype ) throws Exception
    {
        // If node conforms to any stereotype, it has a child <eAnnotations source="appliedStereotypes">
        // Stereotype names are found in <contents > subchildren
        XmiNode annotations = node.getChildByTagName("eAnnotations");
        if( annotations == null )
        {
            log.fine("Node " + node.name() + " has no child 'eAnnotations'");
            return false;
        }
        
        String source = XmiUtil.getAttribute(annotations, "source");
        
        if( source == null || !source.equals("appliedStereotypes") )
        {
            log.fine("Attribute source='appliedStereotypes' not found for node: " + node.name());
            return false;
        }
        
        // Now get all children called <contents>. Each child contains the tags for one stereotype.
        Collection<XmiNode> contentChildren = annotations.collectChildrenByTagName("contents");
        for( XmiNode content: contentChildren )
        {
            String expectedId = profile + "_0:" + profile + "__" + stereotype;
            
            if(! expectedId.equals( XmiUtil.getXmiType(content) ))
                continue;
            
            return true;
        }
        
        // No corresponding child was found. Return false
        return false;
    }
    

    public static boolean isUmlPrimitiveType(XmiNode node)
    {
        return getXmiType(node).equals("uml:PrimitiveType");
    }
}
